#ifndef _LINUX_TTY_H
#define _LINUX_TTY_H

/* MOVE ELSEWHERE IN THIS FILE */
#define TTY_BREAK_CHAR	256
#define TTY_FE	0x0800
#define TTY_PE	0x0400
#define TTY_OE	0x0200

/*
 * 'tty.h' defines some structures used by tty_io.c and some defines.
 */

#include <linux/termios.h>

#include <asm/system.h>

#define NR_CONSOLES	8
#define NR_LDISCS	16
#define MAX_TTYS	256

/* stuff for crted line discipline */
#define MAXLLEN 255		/* maximum line length (excluding newline) */
#define COLUMNS 132		/* maximum screen width we expect */
#define NSYNTAX 3		/* number of syntax classes */



/*
 * These are set up by the setup-routine at boot-time:
 */

struct screen_info {
	unsigned char  orig_x;
	unsigned char  orig_y;
	unsigned char  unused1[2];
	unsigned short orig_video_page;
	unsigned char  orig_video_mode;
	unsigned char  orig_video_cols;
	unsigned short orig_video_ega_ax;
	unsigned short orig_video_ega_bx;
	unsigned short orig_video_ega_cx;
	unsigned char  orig_video_lines;
};

extern struct screen_info screen_info;

#define ORIG_X			(screen_info.orig_x)
#define ORIG_Y			(screen_info.orig_y)
#define ORIG_VIDEO_PAGE		(screen_info.orig_video_page)
#define ORIG_VIDEO_MODE		(screen_info.orig_video_mode)
#define ORIG_VIDEO_COLS 	(screen_info.orig_video_cols)
#define ORIG_VIDEO_EGA_AX	(screen_info.orig_video_ega_ax)
#define ORIG_VIDEO_EGA_BX	(screen_info.orig_video_ega_bx)
#define ORIG_VIDEO_EGA_CX	(screen_info.orig_video_ega_cx)
#define ORIG_VIDEO_LINES	(screen_info.orig_video_lines)

#define VIDEO_TYPE_MDA		0x10	/* Monochrome Text Display	*/
#define VIDEO_TYPE_CGA		0x11	/* CGA Display 			*/
#define VIDEO_TYPE_EGAM		0x20	/* EGA/VGA in Monochrome Mode	*/
#define VIDEO_TYPE_EGAC		0x21	/* EGA/VGA in Color Mode	*/

/*
 * This character is the same as _POSIX_VDISABLE: it cannot be used as
 * a c_cc[] character, but indicates that a particular special character
 * isn't in use (eg VINTR ahs no character etc)
 */
#define __DISABLED_CHAR '\0'

/*
 * See comment for the tty_struct structure before changing
 * TTY_BUF_SIZE.  Actually, there should be different sized tty_queue
 * structures for different purposes.  1024 bytes for the transmit
 * queue is way overkill.  TYT, 9/14/92
 */
#define TTY_BUF_SIZE 1024	/* Must be a power of 2 */

struct tty_queue {
	unsigned long data;
	unsigned long head;
	unsigned long tail;
	struct wait_queue * proc_list;
#if 0 /* obsolete */
	unsigned char buf[TTY_BUF_SIZE];
#else
	unsigned short *buf;
#endif
};

#define IS_A_CONSOLE(min)	(((min) & 0xC0) == 0x00)
#define IS_A_SERIAL(min)	(((min) & 0xC0) == 0x40)
#define IS_A_PTY(min)		((min) & 0x80)
#define IS_A_PTY_MASTER(min)	(((min) & 0xC0) == 0x80)
#define IS_A_PTY_SLAVE(min)	(((min) & 0xC0) == 0xC0)
#define PTY_OTHER(min)		((min) ^ 0x40)

#define INC(a) ((a) = ((a)+1) & (TTY_BUF_SIZE-1))
#define DEC(a) ((a) = ((a)-1) & (TTY_BUF_SIZE-1))
#define EMPTY(a) ((a)->head == (a)->tail)
#define LEFT(a) (((a)->tail-(a)->head-1)&(TTY_BUF_SIZE-1))
#define FIRST(a) ((a)->buf[(a)->tail])
#define SECOND(a) ((a)->buf[(TTY_BUF_SIZE-1)&((a)->tail+1)])
#define ADVANCE(a, n)	((a)->tail = (a)->tail + (n) & (TTY_BUF_SIZE-1))
#define LAST(a) ((a)->buf[(TTY_BUF_SIZE-1)&((a)->head-1)])
#define FULL(a) (!LEFT(a))
#define CHARS(a) (((a)->head-(a)->tail)&(TTY_BUF_SIZE-1))

extern void put_tty_queue(char c, struct tty_queue * queue);
extern int get_tty_queue(struct tty_queue * queue);

#define INTR_CHAR(tty) ((tty)->termios->c_cc[VINTR])
#define QUIT_CHAR(tty) ((tty)->termios->c_cc[VQUIT])
#define ERASE_CHAR(tty) ((tty)->termios->c_cc[VERASE])
#define WERASE_CHAR(tty) ((tty)->termios->c_cc[VWERASE])
#define KILL_CHAR(tty) ((tty)->termios->c_cc[VKILL])
#define EOF_CHAR(tty) ((tty)->termios->c_cc[VEOF])
#define START_CHAR(tty) ((tty)->termios->c_cc[VSTART])
#define STOP_CHAR(tty) ((tty)->termios->c_cc[VSTOP])
#define SUSPEND_CHAR(tty) ((tty)->termios->c_cc[VSUSP])
#define LNEXT_CHAR(tty)	((tty)->termios->c_cc[VLNEXT])
#define REPRINT_CHAR(tty) ((tty)->termios->c_cc[VREPRINT])

#define _L_FLAG(tty,f)	((tty)->termios->c_lflag & f)
#define _I_FLAG(tty,f)	((tty)->termios->c_iflag & f)
#define _O_FLAG(tty,f)	((tty)->termios->c_oflag & f)

#define L_CANON(tty)	_L_FLAG((tty),ICANON)
#define L_ISIG(tty)	_L_FLAG((tty),ISIG)
#define L_ECHO(tty)	_L_FLAG((tty),ECHO)
#define L_ECHOE(tty)	_L_FLAG((tty),ECHOE)
#define L_ECHOK(tty)	_L_FLAG((tty),ECHOK)
#define L_ECHONL(tty)	_L_FLAG((tty),ECHONL)
#define L_ECHOCTL(tty)	_L_FLAG((tty),ECHOCTL)
#define L_ECHOKE(tty)	_L_FLAG((tty),ECHOKE)
#define L_TOSTOP(tty)	_L_FLAG((tty),TOSTOP)
#define L_IEXTEN(tty)	_L_FLAG((tty),IEXTEN)

#define I_UCLC(tty)	_I_FLAG((tty),IUCLC)
#define I_NLCR(tty)	_I_FLAG((tty),INLCR)
#define I_CRNL(tty)	_I_FLAG((tty),ICRNL)
#define I_NOCR(tty)	_I_FLAG((tty),IGNCR)
#define I_IXON(tty)	_I_FLAG((tty),IXON)
#define I_IXANY(tty)	_I_FLAG((tty),IXANY)
#define I_IXOFF(tty)	_I_FLAG((tty),IXOFF)
#define I_STRP(tty)	_I_FLAG((tty),ISTRIP)
#define I_BRKINT(tty)	_I_FLAG((tty),BRKINT)
#define I_IGNBRK(tty)	_I_FLAG((tty),IGNBRK)
#define I_IGNPAR(tty)	_I_FLAG((tty),IGNPAR)
#define I_INPCK(tty)	_I_FLAG((tty),INPCK)
#define I_PARMRK(tty)	_I_FLAG((tty),PARMRK)

#define O_POST(tty)	_O_FLAG((tty),OPOST)
#define O_NLCR(tty)	_O_FLAG((tty),ONLCR)
#define O_CRNL(tty)	_O_FLAG((tty),OCRNL)
#define O_NOCR(tty)	_O_FLAG((tty),ONOCR)
#define O_NLRET(tty)	_O_FLAG((tty),ONLRET)
#define O_LCUC(tty)	_O_FLAG((tty),OLCUC)

#define C_SPEED(tty)	((tty)->termios->c_cflag & CBAUD)
#define C_HUP(tty)	(C_SPEED((tty)) == B0)
#define C_READ(tty)	((tty)->termios->c_cflag & CREAD)

struct tty_struct;	/* Handle ANSI C stupidity */

/*
 * Entry points to the line discipline from above (that is, from system
 * calls).
 */
struct ldisc_above {
	void (*close_ldisc)(struct tty_struct *);
	void (*write_flush)(struct tty_struct *);
	int (*read_chan)(struct tty_struct *, struct file *, char *, int);
	int (*write_chan)(struct tty_struct *, struct file *, char *, int);
	int (*select)(struct inode *, struct file *, int, select_table *);
	int (*ioctl)(struct inode *, struct file *, unsigned, unsigned);
	void (*flush_input)(struct tty_struct *);
	void (*flush_output)(struct tty_struct *);
};

/*
 * Entry points to the line discipline from below (that is, from
 * software interrupt handlers).
 */
struct ldisc_below {
	void (*process_input)(struct tty_struct *);
	void (*hangup)(struct tty_struct *);
	void (*timer)(struct tty_struct *);
};


struct ldisc {
	unsigned long name;
	void (*open_ldisc)(struct tty_struct *);
	struct ldisc_above ldisc_above;
	struct ldisc_below ldisc_below;
};

/*
 * Entry points used to control the device.
 */
struct tty_device {
	void (*open)(struct tty_struct *, struct file *);
	void (*close)(struct tty_struct *, struct file *);
	void (*write)(struct tty_struct *);
	void (*ioctl)(struct tty_struct *, struct file *, unsigned, unsigned);
	void (*throttle)(struct tty_struct *, int);
	void (*new_termios)(struct tty_struct *, struct termios *);
	void (*flush_output)(struct tty_struct *);
	void (*output_ctl)(struct tty_struct *, int);
};

/* From update.c: */
#define ARB 4

struct displine {
      struct displine *next;	/* next line */
      int len;			/* length of line */
      char text[ARB];		/* the text of the line */
};

/*
 * Where all of the state associated with a tty is kept while the tty
 * is open.  Since the termios state should be kept even if the tty
 * has been closed --- for things like the baud rate, etc --- it is
 * not stored here, but rather a pointer to the real state is stored
 * here.  Possible the winsize structure should have the same
 * treatment, but (1) the default 80x24 is usually right and (2) it's
 * most often used by a windowing system, which will set the correct
 * size each time the window is created or resized anyway.
 * IMPORTANT: since this structure is dynamically allocated, it must
 * be no larger than 4096 bytes.  Changing TTY_BUF_SIZE will change
 * the size of this structure, and it needs to be done with care.
 * 						- TYT, 9/14/92
 */
struct tty_struct {
	struct termios *termios;
	int pgrp;
	int session;
	unsigned char stopped:1, status_changed:1, packet:1, lnext:1;
	unsigned char ctrl_status;
	short line;
	int flags;
	int count;
	struct winsize winsize;
	int  (*open)(struct tty_struct * tty, struct file * filp);
	void (*close)(struct tty_struct * tty, struct file * filp);
	void (*write)(struct tty_struct * tty);
	int  (*ioctl)(struct tty_struct *tty, struct file * file,
		    unsigned int cmd, unsigned int arg);
	void (*throttle)(struct tty_struct * tty, int status);
	void (*set_termios)(struct tty_struct *tty, struct termios * old);
	struct tty_struct *next;
	struct tty_struct *link;
	unsigned char *write_data_ptr;
	int write_data_cnt;
	void (*write_data_callback)(void * data);
	void *write_data_arg;
	struct ldisc_above la;
	struct ldisc_below lb;
	struct tty_device dev;
	struct tty_queue read_q;
	struct tty_queue write_q;
	struct tty_queue secondary;
	long endhist;
	char *histarea;
	void *edit;
	int column;
	char readbuf[256];
	int writecol;
	char *write_ipos;

	/* Stuff from ed.c: */
	char curline[MAXLLEN+1];	/* line being edited */
	char *point;			/* location of point */
	char *mark;			/* location of mark */
	char *endcurline;		/* end of line */
#if 0
	char histarea[HISTSIZE];	/* history file */
	long endhist;			/* end of history file */
#endif
	long curhist;			/* location in history file */
	char lasthist[MAXLLEN];		/* line after end of history file */
	int lasthistlen;		/* length of lasthist */
	int histprevlen;		/* length of last entry in history file */
	char killbuffer[MAXLLEN];	/* killed text goes here */
	int killsize;			/* length of killed text */
	int prefixarg;			/* if set, evaluating prefix arg */
	int prefixval;			/* value of prefix argument */
	int prefixsign;			/* sign of prefix argument */
	char cmdcharacter;		/* character that invoked this command */
	unsigned char *curmode;		/* base key table for current mode */
	unsigned char *curtable;	/* current command parsing table */
	unsigned char *keytabs;		/* key tables */
	int keyhash;			/* DEBUG */
	char syntaxtbl[NSYNTAX][16];	/* syntax tables */
	char *edstrings;		/* string table */
	char *abbrevtbl;		/* abbreviation table */
	int (*edrecurse)();		/* function to call when recursive edit done */
	int (*ednextc)();		/* function to call on next character */
#if 0
	struct re_pattern_buffer pbuf;	/* buffer containing last search pattern */
#endif
	int savepatlen;			/* length of saved search pattern */
	char savepat[80];		/* saved search pattern */
	char savecurline[MAXLLEN];	/* saved copy of current line */
	int savecurlinelen;		/* length of savecurline */
	char *savepoint;		/* point saved for recursive edit */
	char *savemark;			/* point saved for recursive edit */
	int saveprefixval;		/* prefixval saved here during recursive edit */
	char editprompt;		/* printed during recursive edit */

	/* Stuff from update.c: */
	int scrnwidth;			/* width of screen */
	int eddisplay;			/* if nonzero, editor display is valid */
	int needbeep;			/* if nonzero, beep on next refresh */
	struct displine *virtscrn;	/* what the screen should look like */
	struct displine *realscrn;	/* what the screen does look like */
	int vslines;			/* number of lines on virtual screen */
	int rslines;			/* number of lines on real screen */
	int virtline;			/* cursor position in virtscrn */
	int virtcol;			/* cursor position in virtscrn */
	struct displine *virtlp;	/* virtscrn entry corresponding to virtlne */
	char *virtptr;			/* &virtlp->text[virtcol] */
	int realline;			/* position of cursor on real screen */
	int realcol;			/* position of cursor on real screen */
	int lastrealline;		/* last real line on the screen */
	int startline;			/* end of prompt */
	int startcol;			/* end of prompt */
	char tcspace[120];		/* strings from termcap are stored here */
	int tc_am;			/* cursor wraps around at right of screen */
	char *tc_bc;			/* backspace */
	char *tc_bl;			/* bell */
	int tc_bs;			/* if set, terminal can backspace */
	char *tc_ce;			/* clear to end of line */
	int tc_co;			/* default width of screen */
	char *tc_do;			/* cursor down one line */
	char *tc_nc;			/* cursor right on column */
	char *tc_up;			/* cursor up one line */
	int tc_xn;			/* a newline is ignored after wrap around */
	char PC;			/* pad character (for termcap routines) */

	/* Stuff from output.c: */
	int outstate;
#if 0
	int column;
#endif
	char outline[COLUMNS];
	char prompt[COLUMNS];
	int promptlen;
	int param[2];
	char strparam[256];
	int nparam;
	char *strparamp;
	char esccmd;
	int *curnumber;
	int promptcode;
	int promptset;
	char lastoutline[COLUMNS];
	int lastoutlinelen;

	/* Recent additions */
	int readtimeout;
	int chartimeout;
	char gottimeout;
	int minread;
	};


/*
 * These are the different types of thottle status which can be sent
 * to the low-level tty driver.  The tty_io.c layer is responsible for
 * notifying the low-level tty driver of the following conditions:
 * secondary queue full, secondary queue available, and read queue
 * available.  The low-level driver must send the read queue full
 * command to itself, if it is interested in that condition.
 *
 * Note that the low-level tty driver may elect to ignore one or both
 * of these conditions; normally, however, it will use ^S/^Q or some
 * sort of hardware flow control to regulate the input to try to avoid
 * overflow.  While the low-level driver is responsible for all
 * receiving flow control, note that the ^S/^Q handling (but not
 * hardware flow control) is handled by the upper layer, in
 * copy_to_cooked.  
 */
#define TTY_THROTTLE_SQ_FULL	1
#define TTY_THROTTLE_SQ_AVAIL	2
#define TTY_THROTTLE_RQ_FULL	3
#define TTY_THROTTLE_RQ_AVAIL	4

/*
 * This defines the low- and high-watermarks for the various conditions.
 * Again, the low-level driver is free to ignore any of these, and has
 * to implement RQ_THREHOLD_LW for itself if it wants it.
 */
#define SQ_THRESHOLD_LW	16
#define SQ_THRESHOLD_HW 768
#define RQ_THRESHOLD_LW 16
#define RQ_THRESHOLD_HW 768

/*
 * These bits are used in the flags field of the tty structure.
 * 
 * So that interrupts won't be able to mess up the queues,
 * copy_to_cooked must be atomic with repect to itself, as must
 * tty->write.  Thus, you must use the inline functions set_bit() and
 * clear_bit() to make things atomic.
 */
#define TTY_WRITE_BUSY 0
#define TTY_READ_BUSY 1
#define TTY_CR_PENDING 2
#define TTY_SQ_THROTTLED 3
#define TTY_RQ_THROTTLED 4
#define TTY_IO_ERROR 5
#define TTY_WRITING 6
#define TTY_WAIT_WRITE 7
#define TTY_READING 8
#define TTY_WAIT_READ 9

#define TTY_WRITE_FLUSH(tty) tty_write_flush((tty))
#define TTY_READ_FLUSH(tty) tty_read_flush((tty))

extern void tty_write_flush(struct tty_struct *);
extern void tty_read_flush(struct tty_struct *);

extern struct tty_struct *tty_table[MAX_TTYS];
extern int tty_check_write[];
extern struct tty_struct * redirect;
extern struct ldisc *ldiscs[NR_LDISCS];
extern int fg_console;
extern unsigned long video_num_columns;
extern unsigned long video_num_lines;
extern struct wait_queue * keypress_wait;

#define TTY_TABLE_IDX(nr)	((nr) ? (nr) : (fg_console+1))
#define TTY_TABLE(nr) 		(tty_table[TTY_TABLE_IDX(nr)])

/*	intr=^C		quit=^|		erase=del	kill=^U
	eof=^D		vtime=\0	vmin=\1		sxtc=\0
	start=^Q	stop=^S		susp=^Z		eol=\0
	reprint=^R	discard=^U	werase=^W	lnext=^V
	eol2=\0
*/
#define INIT_C_CC "\003\034\177\025\004\0\1\0\021\023\032\0\022\017\027\026\0"

extern int tty_debug;
#define TTY_TRACE(args)	do {if (tty_debug) printk args;} while (0)

extern long rs_init(long);
extern long lp_init(long);
extern long con_init(long);
extern long tty_init(long);

extern void flush_input(struct tty_struct * tty);
extern void flush_output(struct tty_struct * tty);
extern void wait_until_sent(struct tty_struct * tty);
extern void copy_to_cooked(struct tty_struct * tty);
extern int tty_add_ldisc(struct ldisc * ldisc);
extern int tty_delete_ldisc(struct ldisc * ldisc);
extern void tty_no_op(struct tty_struct * tty);
extern int tty_ioctl(struct inode *, struct file *, unsigned int, unsigned int);
extern int is_orphaned_pgrp(int pgrp);
extern int is_ignored(int sig);
extern int tty_signal(int sig, struct tty_struct *tty);
extern int kill_pg(int pgrp, int sig, int priv);
extern int kill_sl(int sess, int sig, int priv);
extern void tty_hangup(struct tty_struct * tty);
extern void tty_std_hangup(struct tty_struct * tty);
extern void tty_unhangup(struct file *filp);
extern int hung_up(struct file *);
extern void do_SAK(struct tty_struct *tty);

/* tty write functions */

extern void rs_write(struct tty_struct * tty);
extern void con_write(struct tty_struct * tty);

/* serial.c */

extern int  rs_open(struct tty_struct * tty, struct file * filp);

/* pty.c */

extern int  pty_open(struct tty_struct * tty, struct file * filp);

/* console.c */

extern int con_open(struct tty_struct * tty, struct file * filp);
extern void update_screen(int new_console);
extern void blank_screen(void);
extern void unblank_screen(void);

/* vt.c */

extern int vt_ioctl(struct tty_struct *tty, struct file * file,
		    unsigned int cmd, unsigned int arg);

#endif
