/*
 * Stuff for dealing with tty mode 2.
 */

#include "bind.h"		/* defines NSYNTAX */

#define MAXLLEN 255		/* maximum line length (excluding newline) */
#define COLUMNS 132		/* maximum screen width we expect */


struct unused_tty_struct {
struct tty_struct *tty;		/* the tty structure */

/* Stuff from ed.c: */
char curline[MAXLLEN+1];	/* line being edited */
char *point;			/* location of point */
char *mark;			/* location of mark */
char *endcurline;		/* end of line */
#ifdef notdef
char histarea[HISTSIZE];	/* history file */
long endhist;			/* end of history file */
#endif
long curhist;			/* location in history file */
char lasthist[MAXLLEN];		/* line after end of history file */
int lasthistlen;		/* length of lasthist */
int histprevlen;		/* length of last entry in history file */
char killbuffer[MAXLLEN];	/* killed text goes here */
int killsize;			/* length of killed text */
int prefixarg;			/* if set, evaluating prefix arg */
int prefixval;			/* value of prefix argument */
int prefixsign;			/* sign of prefix argument */
char cmdcharacter;		/* character that invoked this command */
unsigned char *curmode;		/* base key table for current mode */
unsigned char *curtable;	/* current command parsing table */
unsigned char *keytabs;		/* key tables */
int keyhash;			/* DEBUG */
char syntaxtbl[NSYNTAX][16];	/* syntax tables */
char *edstrings;		/* string table */
char *abbrevtbl;		/* abbreviation table */
int (*edrecurse)();		/* function to call when recursive edit done */
int (*ednextc)();		/* function to call on next character */
#ifdef notdef
struct re_pattern_buffer pbuf;	/* buffer containing last search pattern */
#endif
int savedpattern;		/* true if current pattern set */
char savecurline[MAXLLEN];	/* saved copy of current line */
int savecurlinelen;		/* length of savecurline */
char *savepoint;		/* point saved for recursive edit */
char *savemark;			/* point saved for recursive edit */
int saveprefixval;		/* prefixval saved here during recursive edit */
char editprompt;		/* printed during recursive edit */

/* Stuff from update.c: */
int scrnwidth;			/* width of screen */
int eddisplay;			/* if nonzero, editor display is valid */
int needbeep;			/* if nonzero, beep on next refresh */
struct displine *virtscrn;	/* what the screen should look like */
struct displine *realscrn;	/* what the screen does look like */
int vslines;			/* number of lines on virtual screen */
int rslines;			/* number of lines on real screen */
int virtline;			/* cursor position in virtscrn */
int virtcol;			/* cursor position in virtscrn */
struct displine *virtlp;	/* virtscrn entry corresponding to virtlne */
char *virtptr;			/* &virtlp->text[virtcol] */
int realline;			/* position of cursor on real screen */
int realcol;			/* position of cursor on real screen */
int lastrealline;		/* last real line on the screen */
int startline;			/* end of prompt */
int startcol;			/* end of prompt */
char tcspace[120];		/* strings from termcap are stored here */
int tc_am;			/* cursor wraps around at right of screen */
char *tc_bc;			/* backspace */
char *tc_bl;			/* bell */
int tc_bs;			/* if set, terminal can backspace */
char *tc_ce;			/* clear to end of line */
int tc_co;			/* default width of screen */
char *tc_do;			/* cursor down one line */
char *tc_nc;			/* cursor right on column */
char *tc_up;			/* cursor up one line */
int tc_xn;			/* a newline is ignored after wrap around */
char PC;			/* pad character (for termcap routines) */

/* Stuff from output.c: */
int outstate;
int column;
char outline[COLUMNS];
char prompt[COLUMNS];
int promptlen;
int param[2];
char strparam[256];
int nparam;
char *strparamp;
char esccmd;
int *curnumber;
int promptcode;
int promptset;
char lastoutline[COLUMNS];
int lastoutlinelen;
};


void ttyoutc(int, struct tty_struct *);
void outreal(char *, int, struct tty_struct *);
void senddata(char *, int, struct tty_struct *);
void sendsig(int, struct tty_struct *);

void edinit(char *, struct tty_struct *);
void edclose(struct tty_struct *);
void newttychars(void);
void newscrnwidth(struct tty_struct *);
void inchar(char, struct tty_struct *);
void edflush(struct tty_struct *);
void refresh(struct tty_struct *);
void dispoff(int, struct tty_struct *);
void freezedisp(int, struct tty_struct *);
void insertchars(char *, int, struct tty_struct *);
#ifdef notdef
void outchars(char *, int, struct tty_struct *);
#else
void tty_put_char(char, struct tty_struct *);
#endif

void gettermcap(struct tty_struct *);
void movetoend(struct tty_struct *);
