/*
 * Update the display of the line being edited on the screen.
 * Copyright (C) 1989 by Kenneth Almquist.
 */


#include <linux/stddef.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include "tty_ed.h"
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/mm.h>


#ifdef notdef	/* declared in tty_ed.h */
struct displine {
      struct displine *next;	/* next line */
      int len;			/* length of line */
      char text[ARB];		/* text of line */
};
#endif


void virtgoto(int, int, struct tty_struct *);
void virtnewline(struct tty_struct *);
void drawch(int, struct tty_struct *);
void realgoto(int, int, struct tty_struct *);
void copytoreal(int, int, struct tty_struct *);
int growscrn(struct tty_struct *);
void addline(struct displine *, struct displine **);


#define bcopy(from, to, len)	memcpy(to, from, len)
#define malloc(nbytes)	((void *)kmalloc(nbytes, GFP_ATOMIC))
#define free(block)	kfree(block)
#define tputs(string, count, func)	outreal(string, strlen(string), tty)



/*
 * Read the termcap entry for this terminal.
 */

void
gettermcap(struct tty_struct *tty) {
#ifdef notdef
      char *termtype;
      char buffer[1024];
      char *tcp = tty->tcspace;
      char *tc_pc;

      errno = 0;
      if ((termtype = getenv("TERM")) == NULL)
	    badinit("$TERM not set");
      if (tgetent(buffer, termtype) <= 0)
	    badinit("Termcap entry not found");
      tty->tc_am = tgetflag("am");
      tty->tc_bs = tgetflag("bs");
      tty->tc_bc = tgetstr("bc", &tcp);
      tty->tc_bl = tgetstr("bl", &tcp);
      tty->tc_ce = tgetstr("ce", &tcp);
      tty->tc_co = tgetnum("co");
      tty->tc_do = tgetstr("do", &tcp);
      tty->tc_nc = tgetstr("nc", &tcp);
      tc_pc = tgetstr("pc", &tcp);
      tty->tc_up = tgetstr("up", &tcp);
      tty->tc_xn = tgetflag("xn");
      tty->PC = (tc_pc != NULL)? *tc_pc : '\0';
      if (tty->tc_up == NULL)
	    badinit("Atty can only run a terminal with a cursor-up escape sequence");
      tty->scrnwidth = tty->tc_co;
#else
      tty->tc_am = 1;
      tty->tc_bs = 1;
      tty->tc_bc = NULL;
      tty->tc_bl = "\7";
      tty->tc_ce = "\33[K";
      tty->tc_co = 80;
      tty->tc_do = "\033[B";
      tty->tc_nc = "\033[C";
      /* tc_pc = NULL; */
      tty->tc_up = "\033[A";
      tty->tc_xn = 1;
      tty->PC = 0;
      tty->scrnwidth = tty->winsize.ws_col;
      if (tty->scrnwidth == 0)
	    tty->scrnwidth = tty->tc_co;
#endif
}



void
newscrnwidth(struct tty_struct *tty) {
      int width;
      struct displine *lp, *next;
      int disp = tty->eddisplay;

      width = tty->winsize.ws_col;
      if (width == 0)
	    width = tty->tc_co;
      if (width == tty->scrnwidth)
	    return;
      if (disp)
	    dispoff(tty->column, tty);
      for (lp = tty->virtscrn ; lp ; lp = next) {
	    next = lp->next;
	    free(lp);
      }
      tty->virtscrn = NULL;
      for (lp = tty->realscrn ; lp ; lp = next) {
	    next = lp->next;
	    free(lp);
      }
      tty->realscrn = NULL;
      tty->vslines = 0;
      tty->scrnwidth = width;
      if (disp)
	    refresh(tty);
}



void
refresh(struct tty_struct *tty) {
      struct displine *lp;
      char *p;
      int pointline, pointcol;
      static int debugflag = 0;		/*DEBUG*/

      if (tty->virtscrn == NULL && growscrn(tty) < 0)
	    return;
      if (tty->eddisplay == 0) {
	    for (lp = tty->realscrn ; lp != NULL ; lp = lp->next)
		  lp->len = 0;
	    tty->startcol = tty->column % tty->scrnwidth;
	    tty->startline = 0;
	    tty->realcol = tty->startcol;
	    tty->realline = 0;
	    if (tty->column != 0) {
		  bcopy(tty->outline, tty->realscrn->text, tty->realcol);
		  tty->realscrn->len = tty->realcol;
	    }
	    tty->lastrealline = 0;
	    tty->eddisplay = 1;
      }
      bcopy(tty->outline, tty->virtscrn->text, tty->startcol);
      tty->virtscrn->len = tty->startcol;
      tty->virtline = 0;
      tty->virtlp = tty->virtscrn;
      tty->virtcol = tty->startcol;
      tty->virtptr = &tty->virtlp->text[tty->startcol];
      if (tty->promptset) {
	    if (tty->startcol != 0)
		  virtnewline(tty);
	    for (p = tty->prompt ; *p ; p++)
		  drawch(*p, tty);
      }
#ifdef notdef
      /* erase everything following the tty->prompt from the virtual screen */
      virtgoto(tty->startline, tty->startcol, tty);
      lp = &tty->virtscrn[tty->startline];
      lp->len = tty->startcol;
      while ((++lp)->text != NULL)
	    lp->len = 0;
#endif
      if (tty->editprompt)
	    drawch(tty->editprompt, tty);
      pointcol = -1;
      if (L_ECHO(tty)) {
	    for (p = tty->curline ; p < tty->endcurline ; p++) {
		  if (p == tty->point) {
			pointline = tty->virtline;
			pointcol = tty->virtcol;
		  }
		  drawch(*p, tty);
	    }
      }
      if (pointcol < 0) {
	    pointline = tty->virtline;
	    pointcol = tty->virtcol;
      }
      tty->virtlp->len = tty->virtcol;
      copytoreal(pointline, pointcol, tty);
}


/*
 * Copy the virtual screen to the real screen.
 */

void
copytoreal(int pointline, int pointcol, struct tty_struct *tty) {
      char *p;
      char *r;
      int line;
      int len;
      int limit;
      int col;
      struct displine *virtlp;
      struct displine *reallp;
      int endline;

      endline = tty->virtline;
      if (endline < tty->lastrealline)
	    endline = tty->lastrealline;
      virtlp = tty->virtscrn;
      reallp = tty->realscrn;
      for (line = 0 ; line <= endline ; line++) {
	    if (virtlp == NULL || reallp == NULL)
		  panic("tty update");
	    if (line > tty->virtline)
		  virtlp->len = 0;
	    len = virtlp->len;
	    limit = reallp->len;
	    if (limit > len)
		  limit = len;
	    p = virtlp->text;
	    r = reallp->text;
	    for (col = 0 ; col < limit ; col++) {
		  if (*p != *r) {
			if (tty->realline != line || tty->realcol != col)
			      realgoto(line, col, tty);
			ttyoutc(*p, tty);
			tty->realcol++;
			*r = *p;
		  }
		  p++;
		  r++;
	    }
	    for ( ; col < len ; col++) {
		  if (*p != ' ') {
			if (tty->realline != line || tty->realcol != col)
			      realgoto(line, col, tty);
			ttyoutc(*p, tty);
			tty->realcol++;
		  }
		  *r++ = *p++;
	    }
	    limit = reallp->len;
	    if (col < limit) {
		  if (tty->tc_ce == NULL || limit - col == 1) {
			for ( ; col < limit ; col++) {
			      if (*r++ != ' ') {
				    if (tty->realline != line || tty->realcol != col)
					  realgoto(line, col, tty);
				    ttyoutc(' ', tty);
				    tty->realcol++;
			      }
			}
		  } else {
			realgoto(line, col, tty);
			tputs(tty->tc_ce, 1, ttyoutc);
		  }
	    }
	    if (tty->realcol >= tty->scrnwidth) {
		  if (tty->tc_xn || ! tty->tc_am)
			outreal("\r\n", 2, tty);
		  tty->realcol = 0;
		  tty->realline++;
	    }
	    reallp->len = virtlp->len;
	    virtlp = virtlp->next;
	    reallp = reallp->next;
      }
      realgoto(pointline, pointcol, tty);
      if (tty->needbeep) {
	    if (tty->tc_bl)
		  tputs(tty->tc_bl, 1, ttyoutc);
	    tty->needbeep = 0;
      }
      tty->lastrealline = tty->virtline;
}



/*
 * Move to beyond the end of the edited stuff, and turn the display off.
 */

void
movetoend(struct tty_struct *tty) {
      realgoto(tty->lastrealline, 0, tty);
      ttyoutc('\n', tty);
      tty->eddisplay = 0;
      tty->column = 0;
      tty->promptset = 0;
}


/*
 * Output the line being edited to the display, and freeze it there.
 * If the data flag is zero, only freeze the prompt.
 */

void
freezedisp(int data, struct tty_struct *tty) {
      int i;
      struct displine *lp;
      int save_echo;
      struct termios *tio;

      tio = tty->termios;
      save_echo = tio->c_lflag;
      tio->c_lflag &=~ ECHO;
      if (data)
	          tio->c_lflag |= ECHO;		/* WHAT IF ECHO IS OFF? */
      refresh(tty);
      tio->c_lflag = save_echo;
      tty->eddisplay = 0;
      tty->column = tty->realcol;
      lp = tty->realscrn;
      for (i = 0 ; i < tty->realline ; i++)
	    lp = lp->next;
      for (i = 0 ; i < tty->column && i < COLUMNS ; i++) {
	    tty->outline[i] = (i < lp->len)? lp->text[i] : ' ';
      }
      tty->promptset = 0;
}



/*
 * Turn display off.
 */

void
dispoff(int col, struct tty_struct *tty) {
      int line;
      struct displine *lp;

      if (tty->eddisplay == 0)
	    return;
      if (tty->virtscrn) {
	    col = col % tty->scrnwidth;
	    tty->virtscrn->len = col;
	    /* NOTE: the following loop could stop at tty->lastrealline */
	    for (lp = tty->virtscrn->next ; lp ; lp = lp->next) {
		  lp->len = 0;
	    }
	    copytoreal(0, col, tty);
      }
      tty->eddisplay = 0;
}



#ifdef notdef
void
virtgoto(int line, int col, struct tty_struct *tty) {
      struct displine *lp;
      char *p;

      while (tty->vslines <= line) {
	    allocline(&tty->virtscrn[tty->vslines], tty);
	    tty->vslines++;
      }
      lp = &tty->virtscrn[line];
      if (lp->len < col) {
	    p = &lp->text[lp->len];
	    do *p++ = ' ';
	    while (++lp->len < col);
      }
      tty->virtline = line;
      tty->virtcol = col;
      tty->virtptr = &lp->text[col];
      tty->virtlen = &lp->len;
}
#else
void
virtnewline(struct tty_struct *tty) {
      tty->virtcol = 0;
      tty->virtline++;
      if (tty->virtlp) {
	    if (tty->virtlp->next == NULL)
		  growscrn(tty);
	    tty->virtlp = tty->virtlp->next;
      }
      if (tty->virtlp)
	    tty->virtptr = &tty->virtlp->text[0];
      else
	    tty->virtptr = NULL;
}
#endif


void
drawch(int c, struct tty_struct *tty) {	/* was "char c" */
      struct displine *virtlp;

      if (! isprint(c)) {
	    if (c == '\t') {
		  do drawch(' ', tty);
		  while ((tty->virtcol & 07) != 0);
		  return;
	    } else {
		  drawch('^', tty);
		  c ^= 0100;
	    }
      }
      tty->virtcol++;
      if ((virtlp = tty->virtlp) != NULL) {
	    *tty->virtptr++ = c;
	    if (virtlp->len < tty->virtcol)
		  virtlp->len = tty->virtcol;
	    if (tty->virtcol >= tty->scrnwidth) {
		  virtnewline(tty);
	    }
      }
}


/*
 * Goto the specified location on the real screen.  Note that we may
 * scroll the screen, so we must use line feed rather than do for
 * downward motion.
 */

void
realgoto(int line, int col, struct tty_struct *tty) {
      char *text;
      int len;
      struct displine *lp;
      int i;

      while (tty->realline > 0 && tty->realline > line) {
	    tputs(tty->tc_up, 1, ttyoutc);
	    tty->realline--;
      }
      while (tty->realline < line) {
#ifdef notdef
	    tputs(tty->tc_do, 1, ttyoutc);
#else
	    ttyoutc('\n', tty);
	    if (O_NLRET(tty))
		  tty->realcol = 0;
#endif
	    tty->realline++;
      }
      if (col < tty->realcol - col) {
	    ttyoutc('\r', tty);
	    tty->realcol = 0;
      }
      if (tty->realcol > col) {
	    if (tty->tc_bc) {
		  do {
			tputs(tty->tc_bc, 1, ttyoutc);
		  } while (--tty->realcol > col);
	    } else {
		  do {
			ttyoutc('\b', tty);
		  } while (--tty->realcol > col);
	    }
      }
      if (tty->realcol < col) {
	    i = tty->realline;
	    for (lp = tty->realscrn ; --i >= 0 ; lp = lp->next);
	    text = lp->text;
	    len = lp->len;
	    do {
		  ttyoutc(tty->realcol < len? text[tty->realcol] : ' ', tty);
	    } while (++tty->realcol < col);
      }
      if (line == 0 && tty->realline > 0) {
	    tputs(tty->tc_up, 1, ttyoutc);
	    tty->realline--;
      }
}



/*
 * Append one line to both the real and virtual screens.  The screens
 * are both grown at once so we don't have to worry about running out
 * of memory during copytoreal.
 */

int
growscrn(struct tty_struct *tty) {
      struct displine *virtline, *realline;
      int size = offsetof(struct displine, text) + tty->scrnwidth;

      if ((virtline = kmalloc(size, GFP_ATOMIC)) == NULL)
	    return -1;
      if ((realline = kmalloc(size, GFP_ATOMIC)) == NULL) {
	    kfree_s(virtline, size);
	    return -1;
      }
      addline(virtline, &tty->virtscrn);
      addline(realline, &tty->realscrn);
      return 0;
}


void
addline(struct displine *line, struct displine **scrn) {
      struct displine *lp, **lpp;

      for (lpp = scrn ; (lp = *lpp) != NULL ; lpp = &lp->next);
      *lpp = line;
      line->next = NULL;
      line->len = 0;
}
